//
// Amiga IFF ILBM/Anim to BMP Converter
// ------------------------------------
//
// Usage
//	anim2bmp myanim.iff
//
// Outputs an image or image sequence.
//
// Compiles with: 
//	dmd -i anim2bmp.d

import std.stdio;
import std.conv;
import std.File;

import std.string;		// lastIndexOf
import std.bitmanip;	// swapEndian

import bmp;

//
// IFF Structs
//

struct my_iff_header{
	align(1):
	char[4] chunk_id; 	// 	FORM
	uint size;
	char[4] format_id; 	//	ILBM / ANIM
};

struct my_chunk_header{
	align(1):
	char[4] chunk_id; 	//BMHD
	uint size;
};

struct my_bmhd_header{
	align(1):
	ushort w;
	ushort h;
	ushort xorigin;
	ushort yorigin;
	ubyte planes;
	ubyte mask;
	ubyte comp;
	char pad1;
	ushort transclr;
	ubyte xaspect;
	ubyte yaspect;
	ushort pagew;
	ushort pageh;
};

struct my_anhd_chunk{
	align(1):
	ubyte op;
	ubyte mask;
	ushort w;
	ushort h;
	ushort x;
	ushort y;
	uint abstime;
	uint reltime;	// jiffies 1/60th
	ubyte interleave;
	ubyte pad0;
	uint bits;
	ubyte[16] pad;
}

struct rgb{
	ubyte r;
	ubyte g;
	ubyte b;
};


rgb[256] MY_RGB;	//0 - 255

int bmhd_w;
int bmhd_h;
int bmhd_d;

// const int bmhd_frames=5;
int bmhd_fps=4;

ubyte[] output_rgb;

const bool quiet=true;

@safe int needed(int w){
	int c=0; 
	
	while(w>0){
		w-=8;
		c=c+1;
	}
	
	// So row has even number of bytes
	if ((c & 1)==1){c++;}	// %
	
	//writeln("IFF Bytes needed per row: %d",c);
	return c;
}
	
@safe int byte_width(int w){
	int bmp_row_width=w*3;
	
	int b=0;
	while(true){
		b+=4;
		if (b>=bmp_row_width){break;}
	}
	
	//writeln("BMP Bytes needed per row: %d",b);
	return b;
}	
	
auto HAM=false;
auto EHB=false;
	
void main(string[] args){
	
	writeln("Anim2Bmp v0.1 D (27/01/24) by ssjx (http://ssjx.co.uk)");
	writeln("------------------------------------------------------");
		
	if (args.length!=2){
		writeln("Can convert an Amiga IFF Animation to a sequence of Windows BMP files.");
		writeln("Usage: anim2bmp myfile.iff");
		return;
	}
	string fn=args[1];
	
	if (exists(fn)==false){
		writeln("File not found!");
		writeln("Usage: anim2bmp myfile.iff");
		return;
	}
	
	// 
	auto dot=lastIndexOf(fn, '.');
	auto outbase=fn[0..dot];
	writeln("Base:",outbase);
	
	// Main IFF header
	File file = File(fn, "rb");

	byte[] frame_a;
	byte[] frame_b;
	ubyte[] dlta_data;

	my_iff_header[1] iff_hdr;
	my_chunk_header[1] iff_chunk;
	my_bmhd_header[1] iff_bmhd;
	
	int bmhd_comp=0,bmhd_mask=0,bmhd_planes=0;
	int bmpsize=0;
	
	// FORM####ANIM
	file.rawRead(iff_hdr);
	
	int total_size=iff_hdr[0].size.swapEndian;
	int read_size=4;
	
	bool anim=false;
	if (iff_hdr[0].format_id=="ANIM"){writeln("Animation");anim=true;} 					
	if (iff_hdr[0].format_id=="ILBM"){writeln("Image");} 
	
	int f=-1;
	int alt=0; 
	while(true){
		f++;
		bool done=false;
		while(true){
			file.rawRead(iff_chunk);
		
			auto chunk_id=iff_chunk[0].chunk_id;
			int chunk_size=iff_chunk[0].size.swapEndian;
		
			switch(chunk_id){
			case "FORM":
				if (quiet==false){
					writeln("-----------------");
					writeln("Loading Frame ",f);
					writeln("--> ID  : ",chunk_id);
					writeln("--> Size  : ",chunk_size);
				}
				
				char[4] is_ilbm;
				file.rawRead(is_ilbm);
				
				read_size+=(chunk_size+8);
				writeln(read_size," / ",total_size);
			break;
			
			case "ANHD":
				my_anhd_chunk[1] iff_anhd;
				file.rawRead(iff_anhd);
				int frame_comp=iff_anhd[0].op;
				int jiffies=cast(int)(iff_anhd[0].reltime.swapEndian);
				bmhd_fps=jiffies/60;
				
				if (quiet==false){
					writeln("Frame animation header ",chunk_size);
					writeln("--> Compression  : ",frame_comp);
					writeln("--> Jiffies  : ",jiffies);
					writeln("--> FPS  : ",bmhd_fps);
				}
			break;
				
			case "DLTA":
				if (quiet==false){
					writeln("Delta Chunk ",chunk_size);
				}
				
				
				dlta_data.length=chunk_size;
				file.rawRead(dlta_data);
				
				if (alt==0){
					if (decode_delta(dlta_data,5,frame_a)==true){done=true;}
				}else{
					if (decode_delta(dlta_data,5,frame_b)==true){done=true;}
				}
			break;
			
			
			case "BMHD":
				file.rawRead(iff_bmhd);
			
				bmhd_w=iff_bmhd[0].w.swapEndian;	
				bmhd_h=iff_bmhd[0].h.swapEndian;	
				
				bmhd_planes=iff_bmhd[0].planes;
				bmhd_comp=iff_bmhd[0].comp;
				
				output_rgb.length=bmhd_w*bmhd_h*3;
				
				int row_width=needed(bmhd_w);
				bmpsize=row_width*(bmhd_h)*(bmhd_planes);
				
				if (quiet==false){
					writeln("Loading Bitmap Header");
					writeln("--> Size  : ",chunk_size);
					writeln("--> Dimensions  : ",bmhd_w," x ", bmhd_h," x ",iff_bmhd[0].planes);
					// writeln("--> Page  : ",iff_bmhd[0].pagew," x ",iff_bmhd[0].pageh);
					writeln("--> Compression  : ",iff_bmhd[0].comp);
					writeln("--> Mask  : ",iff_bmhd[0].mask);
					writeln("--> Pic Bytes  : ",bmpsize);
				}
				
				bmhd_d=iff_bmhd[0].planes;
				
				frame_a.length=bmpsize;
				frame_b.length=bmpsize;
			break;
			
			case "BODY":
				if (quiet==false){
					writeln("Loading BODY");
					writeln("--> Size  : ",chunk_size);
				}
				
				{
					byte[] body_data;
					body_data.length=chunk_size;
					file.rawRead(body_data);
					
					// Decode into Frame A
					if (decode_body(body_data,frame_a,bmhd_comp)==true){
						// Copy to Frame B (Check this is correct!!)
						frame_b=frame_a;
					}
				}
				
				// So we end on even boundary!
				if ((chunk_size & 1)==1){
					file.seek(1,SEEK_CUR);
				}							
					
				done=true;
			break;
			
			
			case "CAMG":
				if (quiet==false){
					writeln("Loading CAMG");
					writeln("--> Size  : ",chunk_size);
				}
				
				uint[1] vp;
				file.rawRead(vp);
				int viewport=vp[0].swapEndian;
				
				if ((viewport & 0x80)==0x80){EHB=true;}
				if ((viewport & 0x800)==0x800){HAM=true;}
				
				if (quiet==false){
					writeln("--> Viewport ",viewport);
					if (HAM==true){writeln("--> Hold And Modify (HAM)");}
					if (EHB==true){writeln("--> Extra Half Bright");}
				}
			break;
			
			case "CMAP":
				int numcols=chunk_size/3;
				if (quiet==false){
					writeln("Loading Colour Map");
					writeln("--> Size  : ",chunk_size);
					writeln( "--> Colours  : ",numcols);
				}
				
				ubyte[] my_rgb_array;
				my_rgb_array.length=chunk_size;
				file.rawRead(my_rgb_array);
			
				int a=0;
				foreach(c;0..numcols){
					MY_RGB[c].r=my_rgb_array[a];
					MY_RGB[c].g=my_rgb_array[a+1];
					MY_RGB[c].b=my_rgb_array[a+2];
					a+=3;
				}
				
				if (EHB==true){
					a=0;
					foreach(c;0..numcols){
						MY_RGB[32+c].r=my_rgb_array[a]>>1;
						MY_RGB[32+c].g=my_rgb_array[a+1]>>1;
						MY_RGB[32+c].g=my_rgb_array[a+2]>>1;
						a+=3;
					}
				}
			break;
			
			default:
				if (quiet==false){
					writeln("Unknown Chunk! ",chunk_id," Size:",chunk_size);
				}
				file.seek(chunk_size,SEEK_CUR);
				
			break;
			}
			
			if (done==true){break;}
		}
		
		// Convert Amiga planar data to Windows chunky
		if (alt==0){
			cydraw(frame_a, output_rgb, bmhd_w, bmhd_h, bmhd_planes);
		}else{
			cydraw(frame_b, output_rgb, bmhd_w, bmhd_h, bmhd_planes);
		}
		alt=1-alt;
		
		string outfn=outbase~to!string(f);
		save_bmp(bmhd_w,bmhd_h,output_rgb,outfn);
		
		if (read_size>=total_size){break;}
		if (anim==false){break;}
		// if (frame.stop_playback==true){break;}
	}
	
	writeln("Finished IFF conversion!");
}


//
// Decoding
//

bool decode_body(byte[] frame,byte [] iff_data, int method){
	bool success=false;
	int framesize=frame.length;
	
	switch (method){
		case 0:
			if (quiet==false){
				writeln("Body not compressed!");
			}
			
			iff_data[]=frame[];
			success=true;
		break;
				
		case 1:
			if (quiet==false){
				writeln("Decompressing Body!");
			}
			
			uint ps=0;
			uint op=0;
			int exit=0;
			
			ubyte value;
			ubyte outcol;
			
			while(true){
				value=frame[ps];
				ps=ps+1;
				if (ps<framesize){
				
					if (value<=127){
						for(int i=0;i<(value+1);i++){
							iff_data[op]=frame[ps];
							ps=ps+1;
							op=op+1;
						}
					}
						
					if (value==128){
						if (quiet==false){writeln("End byte found! ",ps," / ",framesize);}
						exit=1;
					}
						
					if (value>128){
						outcol=frame[ps];
						ps=ps+1;
					
						for(int i=0;i<((255-value)+2);i++){ 
							iff_data[op]=outcol;
							op=op+1;
						}
					}	
					
				}else{
					if (quiet==false){writeln("Exit due to size ",ps," / ",framesize);}
					exit=2;
				}
			
				if (ps>=framesize){break;}
				if (exit>0){break;}
			}
			
			if (quiet==false){writeln("Uncompressed bytes ",op);}		
			success=true;
		break;
		
		default:
			// Should never reach this...
			writeln("Unknown compression method!");
		break;
	}
	
	return success;
}


uint swp(ubyte a,ubyte b,ubyte c,ubyte d){
	uint ta = a;
	uint tb = b;
	uint tc = c;
	uint td = d;
	
	return ta + (tb << 8)+ (tc << 16)+ (td << 24);
}


byte[] plane_data;//=new byte[];

bool decode_delta(ubyte[] data,int method,byte[] outframe){
	bool success=false;
	// int datasize=data.Length;
	
	immutable int bytes_per_row=needed(bmhd_w);
	// writeln("Bytes per Row ",bytes_per_row);
	
	immutable int plane_size=bytes_per_row*bmhd_h;	//Single plane!
	// writeln("Plane size ",plane_size);
	
	plane_data.length=plane_size*bmhd_d;
	
	// Convert CURRENT interleaved to planes!
	int pos=0;
	
	foreach(d;0..bmhd_d){
		foreach(y;0..bmhd_h){
			foreach(x;0..bytes_per_row){
				plane_data[pos]=outframe[(d*bytes_per_row)+(y*bytes_per_row*bmhd_d)+x];
				pos++;
			}
		}
	}
	// writeln("End Plane at ",pos);
	
	if (method==5){
		int fp=0;
		int[16] ptr;
		for(int i=0;i<ptr.length;i++){
			ptr[i]=swp(data[fp+3],data[fp+2],data[fp+1],data[fp]);
			fp+=4;
			// writeln("Plane ",i," at ",ptr[i]);
		}
		
		foreach(p;0..bmhd_d){
			fp=ptr[p];
			if (fp==0){break;}
			
			int col=0;
			
			
			while(true){
			pos=(p*plane_size)+col;
				int ops=data[fp++];
				// writeln("Plane ",p," Col ",col);
				// writeln("=================");
				if (ops!=0){
					
					foreach(oc;0..ops){
						if (data[fp]==0){
							// Repeat a byte
							int byte_count=data[fp+1];
							// writeln("Repeat ",data[fp+2]," x ",byte_count);
							
							foreach(b;0..byte_count){
								plane_data[pos]=data[fp+2];
								pos+=bytes_per_row;
							}
							
							fp+=3;
						}else{
							// High bit set 1000 0000
							if ((data[fp] & 0x80)==0x80){
								int count=(data[fp] & 0x7f);
								// writeln("Copy next ",count," bytes [",pos,"]");
								fp++;
								// Copy next X bytes
								foreach(b;0..count){
									plane_data[pos]=data[fp++];
									pos+=bytes_per_row;
								}
								
							}else{
								// High bit not set (skip bytes)
								int skip=data[fp++];
								// writeln("Skip next ",skip," bytes");
								pos+=(bytes_per_row*skip);
							}
						}
					}
				}
				
				col++;
				if (col==bytes_per_row){break;}
			}
		}
		// writeln("FP:{0} Size:{1}",fp,data.Length);
		success=true;
	}
	
	// Convert simple planar to interleaved!
	pos=0;
	int offset=0;
	for (int d=0;d<bmhd_d;d++){
		offset=d*bytes_per_row;
		for(int y=0;y<bmhd_h;y++){
			for(int x=0;x<bytes_per_row;x++){
				outframe[offset+(y*bytes_per_row*bmhd_d)+x]=plane_data[pos];
				pos++;
			}
		}
	}
	// writeln("End Plane at {0}",pos);
	return success;
}



@safe void cydraw(const byte[] from,ubyte[] to,int width,int height, int planes){
	uint ina=0;
	uint inb;
	uint inc;
	
	ubyte a,b,c,d,e,f,g,h;
	
	ubyte ar,ag,ab;
	
	uint x,xa;
	
	ubyte prev_r;
	ubyte prev_g;
	ubyte prev_b;
	
	ubyte[8] pix;	// Planes
	ubyte[8] green;	// Planes
	ubyte[8] blue;	// Planes
	
	immutable uint iff_row_width=needed(width);
	immutable uint bmp_row_width=byte_width(width);
	
	uint pos;
	
	foreach(j;0..height){
		inb=ina;
		x=0;
		
		prev_r=0;
		prev_g=0;
		prev_b=0;
		
		foreach(i;0..iff_row_width){
			
			// Clear plane arrays
			pix[]=0;
			green[]=0;
			blue[]=0;
						
			inc=inb;
			if (planes<24){
				foreach(depth;0..planes){
					h=from[inc] & 0x01;
					g=(from[inc] & 0x02)>>1;
					f=(from[inc] & 0x04)>>2;
					e=(from[inc] & 0x08)>>3;
					d=(from[inc] & 0x10)>>4;
					c=(from[inc] & 0x20)>>5;
					b=(from[inc] & 0x40)>>6;
					a=(from[inc] & 0x80)>>7;
					
					pix[0]+=a<<depth;
					pix[1]+=b<<depth;
					pix[2]+=c<<depth;
					pix[3]+=d<<depth;
					pix[4]+=e<<depth;
					pix[5]+=f<<depth;
					pix[6]+=g<<depth;
					pix[7]+=h<<depth;
					inc+=iff_row_width;
				}
			}else{
				foreach(depth;0..8){
					h=from[inc] & 0x01;
					g=(from[inc] & 0x02)>>1;
					f=(from[inc] & 0x04)>>2;
					e=(from[inc] & 0x08)>>3;
					d=(from[inc] & 0x10)>>4;
					c=(from[inc] & 0x20)>>5;
					b=(from[inc] & 0x40)>>6;
					a=(from[inc] & 0x80)>>7;
					// Reusing pix array to hold red
					pix[0]+=a<<depth;
					pix[1]+=b<<depth;
					pix[2]+=c<<depth;
					pix[3]+=d<<depth;
					pix[4]+=e<<depth;
					pix[5]+=f<<depth;
					pix[6]+=g<<depth;
					pix[7]+=h<<depth;
					inc+=iff_row_width;
				}
				
				foreach(depth;8..16){
					h=from[inc] & 0x01;
					g=(from[inc] & 0x02)>>1;
					f=(from[inc] & 0x04)>>2;
					e=(from[inc] & 0x08)>>3;
					d=(from[inc] & 0x10)>>4;
					c=(from[inc] & 0x20)>>5;
					b=(from[inc] & 0x40)>>6;
					a=(from[inc] & 0x80)>>7;
					
					green[0]+=a<<(depth-8);
					green[1]+=b<<(depth-8);
					green[2]+=c<<(depth-8);
					green[3]+=d<<(depth-8);
					green[4]+=e<<(depth-8);
					green[5]+=f<<(depth-8);
					green[6]+=g<<(depth-8);
					green[7]+=h<<(depth-8);
					inc+=iff_row_width;
				}
				
				foreach(depth;16..24){
					h=from[inc] & 0x01;
					g=(from[inc] & 0x02)>>1;
					f=(from[inc] & 0x04)>>2;
					e=(from[inc] & 0x08)>>3;
					d=(from[inc] & 0x10)>>4;
					c=(from[inc] & 0x20)>>5;
					b=(from[inc] & 0x40)>>6;
					a=(from[inc] & 0x80)>>7;
					
					blue[0]+=a<<(depth-16);
					blue[1]+=b<<(depth-16);
					blue[2]+=c<<(depth-16);
					blue[3]+=d<<(depth-16);
					blue[4]+=e<<(depth-16);
					blue[5]+=f<<(depth-16);
					blue[6]+=g<<(depth-16);
					blue[7]+=h<<(depth-16);
					inc+=iff_row_width;
				}
			}
			
			foreach(row;0..8){
				
				if (planes<24){
					ar=MY_RGB[pix[row]].r;
					ag=MY_RGB[pix[row]].g;
					ab=MY_RGB[pix[row]].b;
					
					if (HAM==true){
				
						final switch(planes){
							case 6:
								if ((pix[row] & 0x30)==0x20){ 
									// red
									ar=(pix[row] & 0x0f)<<4;
									ar=ar | (pix[row] & 0x0f);
									ag=prev_g;
									ab=prev_b;
								}
								
								if ((pix[row] & 0x30)==0x30){
									// blue
									ar=prev_r;
									ag=(pix[row] & 0x0f)<<4;
									ag=ag | (pix[row] & 0x0f);
									ab=prev_b;
								}
							
								if ((pix[row] & 0x30)==0x10){ 
									// green
									ar=prev_r;
									ag=prev_g;
									ab=(pix[row] & 0x0f)<<4;
									ab=ab | (pix[row] & 0x0f);
								}
							break;
							
							case 8:
								if ((pix[row] & 0xc0)==128){ 
									// red
									ar=(pix[row] & 0x3f)<<2;
									ag=prev_g;
									ab=prev_b;
								}
								
								if ((pix[row] & 0xc0)==192){
									// blue
									ar=prev_r;
									ag=(pix[row] & 0x3f)<<2;
									ab=prev_b;
								}
							
								if ((pix[row] & 0xc0)==64){ 
									// green
									ar=prev_r;
									ag=prev_g;
									ab=(pix[row] & 0x3f)<<2;
								
								}
							break;
						}
					
						prev_r=ar;
						prev_g=ag;
						prev_b=ab;
					}
					
				}else{
					ar=pix[row];
					ag=green[row];
					ab=blue[row];
				}
				
				// Write pixel
				pos=(((height-j)-1)*bmp_row_width)+(x*3);
				
				to[pos]=ab;
				to[pos+1]=ag;
				to[pos+2]=ar;
				//
				
				x=x+1;
				if (x==width){break;}
			}
				
			inb+=1;
			if (x==width){break;}
		}
		
		ina+=(iff_row_width*planes);
	}
}


	

